#include "ne.h"
#include "hash.h"
#include <syslog.h>

/* The standard macro descriptor allocation dimension. */

#define STD_MACRO_DESC_SIZE	1024


/* The maximum width for a command (used for displaying the command
names with the string requester). It allows four columns on an 80x25 screen.
*/

#define MAX_COMMAND_WIDTH		18


/* This structure represents a command. It includes a long and a short name,
a NULL-terminated vector of help strings (of specified length) and some flags
which are related to the syntax and the semantics of the arguments. */

typedef struct {
	const char *name, *short_name;
	const char * const *help;
	int help_len;
	int flags;
} command;


#define NO_ARGS					(1<<1)			/* This command must be called without argument. */
#define ARG_IS_STRING			(1<<2)			/* The argument is a string (default is a number). */
#define IS_OPTION					(1<<3)			/* The command controls an option,
															and can be played while exec_only_options is true. */
#define DO_NOT_RECORD			(1<<4)			/* Never record this command. */
#define EMPTY_STRING_OK			(1<<5)			/* This command can accept an empty string ("") as an argument. */


/* These macros makes the following vector more readable. */

#define HELP_LEN(x)			(sizeof(x ## _HELP) / sizeof(char *) - 1)
#define NAHL(x) x ## _NAME, x ##_ABBREV, x ## _HELP, HELP_LEN(x)

/* This is the command vector. Note that the command names come from names.h,
   and the help names come from help.h. */

static const command commands[ACTION_COUNT] = {


};


/* Checks whether the command line m starts with the command c. Return 0 on
	success, non-zero on failure. */

int cmdcmp(const unsigned char *c, const unsigned char *m) {

//	assert(c != NULL);
//	assert(m != NULL);
return 0;

	while (*c && ascii_up_case[*c] == ascii_up_case[*m]) {
		c++;
		m++;
	}

	return *c || *m && !isasciispace(*m) ;
}





/* This table *can* have conflicts, so that we can keep its size much
	smaller. */

static macro_desc *macro_hash_table[MACRO_HASH_TABLE_SIZE];


/* This is the command name hashing function.We consider only the 5 least
	significant bits because they are the bits which distinguish characters,
	independently of their case. We are not interested in strings which contain
	non-alphabetical characters, because they will certainly generate an error
	(the only exception notably being "R1"). We should subtract 1 to s[i], but
	this doesn't seem to produce any improvement. hash_macro() act as hash(),
	but uses MACRO_HASH_TABLE_SIZE for its modulo. */


static int hash(const unsigned char * const s, int len) {
	int h = -1;

	while(len-- != 0) h = (h * 31 + ascii_up_case[s[len]]) % HASH_TABLE_SIZE;
	return ( h + HASH_TABLE_SIZE ) % HASH_TABLE_SIZE;
}


static int hash_macro(const unsigned char * const s, int len) {
	int h = -1;

	while(len-- != 0) h = (h * 31 + ascii_up_case[s[len]]) % MACRO_HASH_TABLE_SIZE;
	return ( h + MACRO_HASH_TABLE_SIZE ) % MACRO_HASH_TABLE_SIZE;
}



/* Parses a command line. This function has an interface which is slightly
	varied with respect to the other functions of ne. In case of a parsing
	error, an error index *with sign inverted* is passed back. In case parsing
	succeeds, an (greater or equal to zero) action is returned, and the
	numerical or string argument is passed in the variables pointed to by
	num_arg or string_arg, respectively, if they are non-NULL. Otherwise, the
	argument is not passed back. The string argument is free()able. -1 and NULL
	denote the lack of an optional numerical or string argument, respectively.
	NOP is returned on a NOP command, or on a comment line (any line whose first
	non-space character is a non alphabetic character). Note that the various
	syntax flags are used here. */

action parse_command_line(const unsigned char * command_line, int * const num_arg, unsigned char ** const string_arg, const int exec_only_options) {

	int h;
	action a;
	const unsigned char *p = command_line;

	if (num_arg) *num_arg = -1;
	if (string_arg) *string_arg = NULL;

	if (!command_line) return NOP_A;

	if (!*p) return NOP_A;

	while(isasciispace(*p)) p++;

	command_line = p;

	if (!isalpha(*p)) return NOP_A;

	while(*p && !isasciispace(*p)) p++;

	h = hash(command_line, p - command_line);

	if ((a = hash_table[h]) && !cmdcmp(commands[--a].name, command_line)
		|| (a = short_hash_table[h]) && !cmdcmp(commands[--a].short_name, command_line)) {

		while(isasciispace(*p)) p++;

		if (!(*p && (commands[a].flags & NO_ARGS))) {
			if (!*p || (commands[a].flags & ARG_IS_STRING) || isxdigit(*p) || *p == 'x' || *p =='X') {
				if ((commands[a].flags & IS_OPTION) || !exec_only_options) {
					if (*p) {
						if ((commands[a].flags & ARG_IS_STRING) && string_arg) {
							int len = strlen(p);

							if (len > 1 && *p == '"' && p[len - 1] == '"') {
								p++;
								len -= 2;
							}

							if (len == 0 && !(commands[a].flags & EMPTY_STRING_OK)) return -STRING_IS_EMPTY;

							if (!(*string_arg = malloc(len + 1))) return -OUT_OF_MEMORY;
							memcpy(*string_arg, p, len);
							(*string_arg)[len] = 0;
						}
						else if (num_arg) {
							char *q;
							*num_arg = strtol(p, &q, 0);
							if (*q && !isasciispace(*q)) return -NOT_A_NUMBER;
						}
					}
					return a;
				}
				D(fprintf(stderr,"parse_command error: Can execute only options.\n");)
				return -CAN_EXECUTE_ONLY_OPTIONS;
			}
			D(fprintf(stderr,"parse_command error: Has numeric argument.\n");)
			return -HAS_NUMERIC_ARGUMENT;
		}
		D(fprintf(stderr,"parse_command error: Has no argument.\n");)
		return -HAS_NO_ARGUMENT;
	}
   D(fprintf(stderr,"parse_command error: No such command.\n");)
	return -NO_SUCH_COMMAND;
}


/* Parses and executes a command line. Standard error codes are returned. If
	the search for a standard command fails, we try to execute a macro in ~/.ne
	with the same name. */

int execute_command_line(buffer *b, const unsigned char *command_line) {

	int n, a, len = strlen(command_line);
	encoding_type encoding = detect_encoding(command_line, len);
	unsigned char *p;

	if (b->encoding != ENC_ASCII && encoding != ENC_ASCII && b->encoding != encoding) return INCOMPATIBLE_COMMAND_ENCODING;

	if ((a = parse_command_line(command_line, &n, &p, b->exec_only_options)) >= 0) 
	{
		return do_action(b, a, n, p);
	}

	a = -a;

	if ((a == NO_SUCH_COMMAND) && (a = execute_macro(b, command_line)) == CANT_OPEN_MACRO) a = NO_SUCH_COMMAND;

	return a;
}



/* Allocates a macro descriptor. It does not allocate the internal character
	stream, which has to be allocated and stuffed in separately. */

macro_desc *alloc_macro_desc(void) {
	return calloc(1, sizeof(macro_desc));
}



/* Frees a macro descriptors. */

void free_macro_desc(macro_desc *md) {

	if (!md) return;

	assert_macro_desc(md);

	free(md->name);
	free_char_stream(md->cs);
	free(md);
}



/* Here we record an action in a character stream. The action name is expanded
	in a short or long name, depending on the value of the verbose parameter.  A
	numerical or string argument are expanded and copied, too. If the command
	should not be recorded (for instance, ESCAPE_A) we return. */

void record_action(char_stream *cs, action a, int c, unsigned char *p, int verbose) {

	char t[MAX_INT_LEN + 2];

	if (commands[a].flags & DO_NOT_RECORD) return;

	if (verbose) add_to_stream(cs, commands[a].name, strlen(commands[a].name));
	else add_to_stream(cs, commands[a].short_name, strlen(commands[a].short_name));


	if (c >= 0) {
		sprintf(t, " %d", c);
		add_to_stream(cs, t, strlen(t));
	}
	else if (p) {
		add_to_stream(cs, " ", 1);
		if (!*p || isasciispace(*p)) add_to_stream(cs, "\"", 1);
		add_to_stream(cs, p, strlen(p));
		if (!*p || isasciispace(*p)) add_to_stream(cs, "\"", 1);
	}

	add_to_stream(cs, "", 1);

}


/* A support function for optimize_macro(). It examines a string to see if it
	is a valid "InsertChar ##" command. If it is, then insertchar_val() returns
	the character code, otherwise it returns 0. */

static int insertchar_val( const unsigned char *p ) {
	int h;
	const unsigned char *cmd;
	action a;
    
	if ( !p || !*p ) return 0;
    
	while(isasciispace(*p)) p++;

	cmd = p;
	
	if (!isalpha(*p)) return 0;
	
	while(*p && !isasciispace(*p)) p++;
	
	h = hash(cmd, p - cmd);
	
	if (((a = hash_table[h]) && !cmdcmp(commands[--a].name, cmd)
		|| (a = short_hash_table[h]) && !cmdcmp(commands[--a].short_name, cmd))	&& a == INSERTCHAR_A) {

		while(isasciispace(*p)) p++;
		h = strtol(p, (char **)&cmd, 0);
		return *cmd || h < 0 ? 0 : h;
	}
	return 0;
}

/* Looks through the macro stream for consecutive runs of InsertChar commands
	and replaces them with appropriate InsertString commands. This makes macros
	much easier to read if and when they have to be edited. Note that if the
	character inserted by InsertChar is not an ASCII character, then we should
	leave it as an InsertChar command to maximize portability of the macros. */

void optimize_macro( char_stream *cs, int verbose ) {
	char *cmd;
	int	pos;
	int	chr;
	int	building = 0;
	 
	if ( !cs || !cs->len ) return;

	for (pos = 0; pos < cs->len; pos += strlen(&cs->stream[pos]) + 1 ) {
		cmd = &cs->stream[pos];
		chr = insertchar_val( cmd );
		if (chr < 0x80 && isprint(chr)) {
			char two[2] = " ";
			two[0] = chr;
			delete_from_stream( cs, pos, strlen(cmd) + 1 );
			if ( building ) {
				building++;
				insert_in_stream( cs, two, building, 1 );
				}
			else {
				const char *insert;
				int	len;
				insert = verbose ? INSERTSTRING_NAME : INSERTSTRING_ABBREV;
				len	 = strlen(insert);
				insert_in_stream( cs, "\"",	pos, 2 );	/* Closing quote */
				insert_in_stream( cs, two,	 pos, 1 );		/* The character itself */
				insert_in_stream( cs, " \"",  pos, 2 );	/* space and openning quote */
				insert_in_stream( cs, insert, pos, len );	/* The command itself */
				building = pos + len + 2;						/* This is where the char is now */
			}
		}
		else {
			building = 0;
		}
	}
}
  

/* This function is the ultimate goal of this file. It plays a character
	stream, considering each line as a command line. It polls the global stop
	variable in order to check for the user interrupting. Note that the macro is
	duplicated before execution: this is absolutely necessary, for otherwise a
	call to CloseDoc, Record or UnloadMacros could free() the block of memory
	which we are executing. */

int play_macro(buffer *b, char_stream *cs) {

	int error = OK, len;
	unsigned char *p, *stream;

	if (!cs) return ERROR;

	/* If len is 0 or 1, the character stream does not contain anything. */

	if ((len = cs->len) < 2) return OK;

	if (!(p = stream = malloc(len))) return OUT_OF_MEMORY;

	memcpy(stream, cs->stream, len);

	stop = FALSE;

	while(!stop && p - stream < len) {	
#ifdef TEST
		fprintf(stderr, "%s\n", p); /* During tests, we output to stderr the current command. */
#endif

		if (error = execute_command_line(b, p))
#ifndef TEST
			break /* During tests, we never interrupt a macro. */
#endif
				;
		p += strlen(p) + 1;
	}

	free(stream);

	return stop ? STOPPED : error;
}



/* Loads a macro, and puts it in the global macro hash table.  file_part is
	applied to the name argument before storing it and hashing it.  Note that if
	the macro can't be opened, we retry prefixing its name with the preferences
	directory name (~/.ne/). Thus, for instance, all autopreferences file whose
	name does not conflict with internal commands can be executed transparently
	just by typing their name. */


macro_desc *load_macro(const char *name) {

	int h;
	char *macro_dir, *prefs_dir;
	char_stream *cs;
	macro_desc *md, **m;

	assert(name != NULL);

	if (!(md = alloc_macro_desc())) return NULL;

	cs = load_stream(md->cs, name, FALSE, FALSE);

	if (!cs && (prefs_dir = exists_prefs_dir()) && (macro_dir = malloc(strlen(prefs_dir) + 2 + strlen(name)))) {
		strcat(strcpy(macro_dir, prefs_dir), name);
		cs = load_stream(md->cs, macro_dir, FALSE, FALSE);
		free(macro_dir);
	}

	if (!cs && (prefs_dir = exists_gprefs_dir()) && (macro_dir = malloc(strlen(prefs_dir) + 2 + strlen(name)))) {
		strcat(strcpy(macro_dir, prefs_dir), name);
		cs = load_stream(md->cs, macro_dir, FALSE, FALSE);
		free(macro_dir);
	}

	if (cs) {
		md->cs = cs;
		md->name = str_dup(file_part(name));

		h = hash_macro(md->name, strlen(md->name));

		m = &macro_hash_table[h];

		while(*m) m = &((*m)->next);

		*m = md;

		return md;
	}

	free_macro_desc(md);
	return NULL;
}



/* Executes a named macro. If the macro is not in the global
macro list, it is loaded. A standard error code is returned. */

int execute_macro(buffer *b, const char *name) {

	const char *p;
	macro_desc *md;
	int h;

	p = file_part(name);

	h = hash_macro(p, strlen(p));

	md = macro_hash_table[h];

	while(md && cmdcmp(md->name, p)) md = md->next;

	if (!md) md = load_macro(name);

	assert_macro_desc(md);

	if (md) return play_macro(b, md->cs);

	return CANT_OPEN_MACRO;
}


/* Clears up the macro table. */

void unload_macros(void) {

	int i;
	macro_desc *m, *n;

	for(i = 0; i < MACRO_HASH_TABLE_SIZE; i++) {

		m = macro_hash_table[i];
		macro_hash_table[i] = NULL;

		while(m) {
			n = m->next;
			free_macro_desc(m);
			m = n;
		}
	}
}



/* This function helps. The help text relative to the command name pointed to
	by p is displayed (p can also contain arguments). If p is NULL, the
	alphabetically ordered list of commands is displayed with the string
	requester. The help finishes when the user escapes. */

void help(char *p) {

	action a;
	int i;
	D(fprintf(stderr,"Help Called with parm %ld.\n",p);)
	do {
		if (p || (i = request_strings(command_names, ACTION_COUNT, MAX_COMMAND_WIDTH)) >= 0) {
         D(fprintf(stderr,"Help check #2: p=%lx, i=%d\n",p,i);)
			if (p) {
				for(i = 0; i < strlen(p); i++) if (isasciispace((unsigned char)p[i])) break;

				i = hash(p, i);

				if ((a = hash_table[i]) && !cmdcmp(commands[--a].name, p)
				|| (a = short_hash_table[i]) && !cmdcmp(commands[--a].short_name, p)) i = a;
				else i = -1;

				p = NULL;
			}
			else {
				D(fprintf(stderr,"Gonna parse_command_line(\"%s\",NULL,NULL,FALSE);\n",command_names[i]);)
				i = parse_command_line(command_names[i], NULL, NULL, FALSE);
				D(fprintf(stderr,"...and got i=%d\n",i);)
			}
			
			if (i < 0) {
				i = 0;
				continue;
			}

			assert(i >= 0 && i < ACTION_COUNT);

			i = request_strings(commands[i].help, commands[i].help_len, ne_columns);
		}
	} while(i >= 0);
}
