#include <sys/types.h>		/* pid_t */
#include <errno.h>			/* errno */
#include <signal.h>			/* sigaction() */
#include <stdio.h>			/* puts() */
#include <stdlib.h>			/* exit() */
#include <string.h>			/* strcmp() */
#include <unistd.h>			/* fork() */

/* waitpid() */
#ifdef HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(status) ((unsigned)(status) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(status) (((status) & 255) == 0)
#endif

#include "ice.h"

//extern int errno;

static EXTRA_LINE el_leave[1] = {};
static INPUTLINE il_filt;
static PANEL_DESC pd_hist =   { 0,0,0,-1,0,0,el_leave,&il_filt,0 };
PANEL_DESC *panel = 0;
PANEL_HIST panel_hist = { &pd_hist };
DISPLAY display = { 0,0,0,0,0,0,0 };
TERM_DATA term_data;
TEXTLINE *textline = 0, line_cmd, line_tmp, line_dir;
const void *pcfg[CFG_VARIABLES];

#define MAX_SHELL_ARGS	8
static char *shell_argv[MAX_SHELL_ARGS + 2 + 1];
static int cmd_index;		/* which parameter is the command
							   to be executed */

static int shelltype(const char *shell)
{
	size_t len;

	shell = (char *)base_name(shell);
	len = strlen(shell);
	if (len >= 2 && strcmp(shell + len - 2,"sh") == 0)
		return (len >= 3 && shell[len - 3] == 'c');
	return 2;
}

static int parse_shellprog(const char *shellprog)
{
	char *pch;
	int i, ac;
	FLAG arg_end;
	static USTRING dup = { 0,0 };

	us_copy(&dup,shellprog);

	/* split to arguments */
	for (arg_end = 1, i = 0, pch = USTR(dup); *pch; pch++ )
		if (*pch == ' ') {
			*pch = '\0';
			arg_end = 1;
		}
		else if (TCLR(arg_end)) {
			if (i >= MAX_SHELL_ARGS) {
//				txt_printf(
//				  "CONFIG: Incorrect SHELLPROG: too many arguments\n"
//				  "        limit is " STR(MAX_SHELL_ARGS) "\n");
				return -1;
			}
			shell_argv[i++] = pch;
		}
	shell_argv[ac = i] = 0;

	for (cmd_index = 0, i = 1; i < ac; i++)
		if (strcmp(shell_argv[i],"<COMMAND>") == 0) {
			cmd_index = i;	/* real command belongs here */
			break;
		}

	/* append: -c <COMMAND> */
	if (cmd_index == 0) {
		shell_argv[ac] = "-c";
		cmd_index = ac + 1;
		shell_argv[ac + 2] = 0;
	}

	term_data.shelltype = shelltype(shell_argv[0]);

	return 0;
}



void exec_shell_reconfig()
{
	parse_shellprog(term_data.shell);
}

/* prepare to execute commands according to "SHELLPROG" */

void exec_prompt_reconfig()
{
	const char *prompt, *chars;
	static USTRING cmd_prompt = { 0,0 };
	static char prompt_char[4] = " ? ";

	prompt = config_str(CFG_PROMPT);

	if (prompt[0] == '\0') {
		/* PROMPT=AUTO */
		prompt = (char *)base_name(shell_argv[0]);
		switch (term_data.shelltype) {
		case 0:
			chars = "$#";	/* Bourne shell or similar */
			break;
		case 1:
			chars = "%#";	/* C-shell or similar */
			break;
		default:
			chars = ">>";	/* some other shell */
			break;
		}
		prompt_char[1] = chars[term_data.isroot];
	}
	else
		prompt_char[1] = '\0';

	us_cat(&cmd_prompt,term_data.isroot ? "ROOT " : "",prompt,prompt_char,(char *)0);
}

void exec_initialize(void)
{
	exec_shell_reconfig();
	exec_prompt_reconfig();	/* prompt needs to know the shell */
}




static int execute(const char *command)
{
	pid_t childpid;
	FLAG failed;
	int status, code;
	struct sigaction act;

	failed = 1;
	xterm_title_set(1,command);
	childpid = fork();
	if (childpid == -1)
		printf("EXEC: Cannot create new process (%s)\n",
		  strerror(errno));
	else if (childpid == 0) {
		/* child process = command */
#ifdef _POSIX_JOB_CONTROL
		/* move this process to a new foreground process group */
		childpid = getpid();
		setpgid(childpid,childpid);
		tcsetpgrp(STDIN_FILENO,childpid);
#endif

		/* reset signal dispositions */
		act.sa_handler = SIG_DFL;
		act.sa_flags = 0;
		sigemptyset(&act.sa_mask);
		sigaction(SIGINT,&act,0);
		sigaction(SIGQUIT,&act,0);
#ifdef _POSIX_JOB_CONTROL
		sigaction(SIGTSTP,&act,0);
		sigaction(SIGTTIN,&act,0);
		sigaction(SIGTTOU,&act,0);
#endif

		/* execute the command */
		((const char **)shell_argv)[cmd_index] = command;
		execv(shell_argv[0],shell_argv);
		printf("EXEC: Cannot execute shell %s (%s)\n",
		  shell_argv[0],strerror(errno));
		exit(99);
		/* NOTREACHED */
	}
	else {
		/* parent process = Ice */
#ifdef _POSIX_JOB_CONTROL
		/* move child process to a new foreground process group */
		setpgid(childpid,childpid);
		tcsetpgrp(STDIN_FILENO,childpid);
#endif
		for (; /* until break */;) {
			/* wait for child process exit or stop */
			while (waitpid(childpid,&status,WUNTRACED) < 0)
				/* ignore EINTR */;
#ifdef _POSIX_JOB_CONTROL
			/* move Ice to foreground */
			tcsetpgrp(STDIN_FILENO,term_data.pid);
#endif
			if (!WIFSTOPPED(status))
				break;

			puts(
				"\r\n"
				"\r\n"
				"EXEC: Ice does not provide job control\r\n"
				"EXEC: suspended command will be resumed shortly\r\n"
				"\r\n");
			fflush(stdout);
#ifdef _POSIX_JOB_CONTROL
			/* move command back to foreground */
			tcsetpgrp(STDIN_FILENO,childpid);
#endif
			sleep (2);
			kill(-childpid,SIGCONT);
		}

		tty_reset();
		if (WIFEXITED(status)) {
			code = WEXITSTATUS(status);
			if (code == 0) {
				failed = 0;
				//fputs("\nCommand successful. ",stdout);
				//fputs("\n",stdout);
			}
			//else printf("\nExit code = %d. ",code);
		}
		else {
			code = WTERMSIG(status);
			printf("\nAbnormal termination, signal %d",code);
#ifdef HAVE_STRSIGNAL
			signame = strsignal(code);
			if (signame != 0)
				printf(" (%s)",signame);
#else
#if HAVE_DECL_SYS_SIGLIST
			signame = sys_siglist[code];
			if (signame != 0)
				printf(" (%s)",signame);
#endif
#endif

#ifdef WCOREDUMP
			if (WCOREDUMP(status))
				fputs(", core image dumped",stdout);
#endif
			putchar('\n');
		}
	}

	xterm_title_set(0,command);
	tty_press_enter();
	xterm_title_set(0,0);

	return failed;
}



/*
 * function returns 1 if the command has been executed
 * (successfully or not), otherwise 0
 */
int execute_cmd(const char *cmd)
{
	int do_exec;
/*
	if ( (dir = check_cd(cmd)) ) {
		hist_save(cmd,0);
		win_heading();
		win_panel();
		win_remark("directory changed");
		return 1;
	}
*/
	curses_stop();
//	putchar('\n');
	puts(cmd);
//	putchar('\n');
	fflush(stdout);

//	warn_level = print_warnings(cmd);
	do_exec = 0 ;
//	if (do_exec)
	hist_save(cmd,execute(cmd) != 0);

	curses_restart();
	//if (warn_level >= 2 && TCLR(hint))		win_remark("warnings can be turned off in the config panel");

	return do_exec;
}
