#include <sys/types.h>			
#include <sys/stat.h>			/* fstat() */
#include <ctype.h>				/* tolower */
#include <fcntl.h>				/* open() */
#include <stdio.h>				/* sprintf() */
#include <stdlib.h>				/* malloc() */
#include <string.h>				/* strlen() */
#include <unistd.h>				/* read() */
#include <limits.h>				/* SSIZE_MAX */
#include "ice.h"

//extern int errno;

static const char *user_config_file; /* personal cfg filename */

typedef struct {
	CODE code;					/* CFG_XXX */
	const char *extra_val;		/* if defined - name of special value
								   (represented as 0) */
	int min, max, initial, current, new;	/* values */
	const char *desc[4];		/* if defined - show this text instead
								   of numbers (enumerated type) */
} CNUM;

static CNUM table_numeric[] = {
	/* enumerated */
	{ 1,	0, 2, 4, 2, 0, 0,
		{	"2 screen lines",
			"3 screen lines",
			"4 screen lines" } },
	{ 9,	0, 0, 1, 0, 0, 0,
		{	"1 KiB is 1024 bytes (IEC standard)",
			"1 KB is 1000 bytes (SI standard)" } },
	{ 10,	0, 0, 2, 0, 0, 0,
		{	"AUTO (according to the current locale)",
			"1.000.000 (dot separated)",
			"1,000,000 (comma separated)" } },
	{ 11,		0, 0, 2, 0, 0, 0,
		{	"AUTO (according to the current locale)",
			"12 hour clock",
			"24 hour clock" } },
	{ 13,	0, 0, 1, 1, 0, 0,
		{	"byte order (based on the character codes)",
			"dictionary order (based on the human language)" } },
	{ 0,		0, 0, 2, 0, 0, 0,
		{	"-----",
			"=====",
			"line graphics (not supported on some terminals)" } },
	{ 3,	0, 0, 2, 2, 0, 0,
		{	"Do not group",
			"Group: directories, special files, plain files",
			"Group: directories, devices, special files, plain files" }
	},
	{ CFG_LAYOUT,		0, 0, 2, 0, 0, 0,
		{	"Layout #1",
			"Layout #2",
			"Layout #3" } },
	{ 25,		0, 0, 1, 1, 0, 0, { "No", "Yes" } },
	{ 26,	0, 0, 1, 1, 0, 0, { "No", "Yes" } },
	{ 27,	0, 0, 1, 1, 0, 0, { "No", "Yes" } },
	{ CFG_XTERM_TITLE,	0, 0, 2, 1, 0, 0,
		{	"No",
			"AUTO (checking the terminal type $TERM)",
			"Yes" } },
	/* really numeric */
	{ 31,		"AUTO",	10, 100,  0, 0, 0, { 0 } },
	{ 32,		"AUTO",	10, 100,  0, 0, 0, { 0 } },
	{ 33,		0,		10, 100, 40, 0, 0, { 0 } }
};

typedef struct {
	CODE code;					/* CFG_XXX */
	const char *extra_val;		/* if defined - name of special value
								   (represented as "") */
	char *initial;
	char current[60 + 1], new[60 + 1];
} CSTR;

static CSTR table_string[] = {
	{ 15,		0,		"more $f",		"", "" },
	{ 16,		0,		"vi $f",		"", "" },
	{ 17,		0,		"cp -ir $f $2",	"", "" },
	{ 18,		0,		"mv -i $f $2",	"", "" },
	{ 19,		0,		"mkdir ",		"", "" },
	{ 20,		0,		"rm $f",		"", "" },
	{ 21,		0,		"lpr $f",		"", "" },
	{ 22,		0,		"",				"", "" },
	{ 23,		0,		"",				"", "" },
	{ 24,		0,		"",				"", "" },
	{ 28,			"HOME",	"",				"", "" },
	{ 12,		"AUTO",	"",				"",	"" },
	{ 29,		"NONE",	"",				"", "" },
	{ 5,		0, "$d $S $>$t $M $*|  $p $o $L", "", "" },
	{ 6,		0, "$d $S $t $*|  $p $o", "", "" },
	{ 7,		0,
	  "$p $o $s $d $>$t $*|  mode=$m atime=$a ctime=$i links=$l", "", "" },
	{ CFG_PROMPT,		"AUTO",	"",				"", "" },
	{ 30,		0,		"",				"", "" },
	{ CFG_SHELLPROG,	"AUTO",	"",				"", "" }
};

static CONFIG_ENTRY config[CFG_VARIABLES] = {
	{ "H_PANEL_SIZE",	0,0,0,0,0 }
};	

/* 'move' values MOV_X2Y understood by set_value() */
#define MOV_SRC_CURRENT	 1
#define MOV_SRC_INITIAL	 2
#define MOV_SRC_NEW		 4
#define MOV_DST_CURRENT	 8
#define MOV_DST_NEW		16
#define MOV_N2C	(MOV_SRC_NEW     | MOV_DST_CURRENT)
#define MOV_C2N	(MOV_SRC_CURRENT | MOV_DST_NEW)
#define MOV_I2N	(MOV_SRC_INITIAL | MOV_DST_NEW)
#define MOV_I2C	(MOV_SRC_INITIAL | MOV_DST_CURRENT)

void err_exit(const char *format, ...)
{
	va_list argptr=NULL;

	/*
	 * all cleanup functions used here:
	 *  - must not call err_exit()
	 *  - must not require initialization
	 */
	xterm_title_restore();
	if (display.curses)
		curses_stop();
	tty_reset();

	fputs("\nTerminating Ice: ",stdout);
	//va_start(argptr,format);
	vprintf(format,argptr);
	//va_end(argptr);
	putchar('\n');

	tty_pgrp_reset();
	exit(0);
	/* NOTREACHED */
}


static void set_value(int code, int move)
{
	int src_num, *dst_num;
	const char *src_str;
	char *dst_str;
	CNUM *pnum;
	CSTR *pstr;

	if (config[code].isnum) {
		pnum = config[code].table;
		if (move & MOV_SRC_CURRENT)
			src_num = pnum->current;
		else if (move & MOV_SRC_INITIAL)
			src_num = pnum->initial;
		else
			src_num = pnum->new;
		if (move & MOV_DST_CURRENT)
			dst_num = &pnum->current;
		else
			dst_num = &pnum->new;
		*dst_num = src_num;
	}
	else {
		pstr = config[code].table;
		if (move & MOV_SRC_CURRENT)
			src_str = pstr->current;
		else if (move & MOV_SRC_INITIAL)
			src_str = pstr->initial;
		else
			src_str = pstr->new;
		if (move & MOV_DST_CURRENT)
			dst_str = pstr->current;
		else
			dst_str = pstr->new;
		strcpy(dst_str,src_str);
	}
}

static CNUM * get_numeric(int code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(table_numeric); i++)
		if (table_numeric[i].code == code)
			return table_numeric + i;

	return 0;
}

static CSTR * get_string(int code)
{
	int i;

	for (i = 0; i < ARRAY_SIZE(table_string); i++)
		if (table_string[i].code == code)
			return table_string + i;
	return 0;
}


void config_initialize(void)
{
	int i;
	CNUM *pnum;
	CSTR *pstr;

	/* initialize 'config' & 'pcfg' tables */
	for (i = 0; i < CFG_VARIABLES; i++) {
		if ( (pnum = get_numeric(i)) ) {
			config[i].isnum = 1;
			config[i].table = pnum;
			pcfg[i] = &pnum->current;
		}
		else if ( (pstr = get_string(i)) ) {
			config[i].table = pstr;
			pcfg[i] = &pstr->current;
		}
		else
			err_exit("BUG: config variable not defined (code %d)",i);
	}

	{
		/* user's configuration file name */
		pathname_set_directory(term_data.homedir);
		user_config_file = estrdup(pathname_join("ice.cfg"));

		for (i = 0; i < CFG_VARIABLES; i++)
			set_value(i,MOV_I2C);
	}

	//panel_cfg.config = config;
}



static USTRING sf_buff = { 0,0 };
static size_t sf_dirlen;



void files_enter(char * buf)
{


	if (execute_cmd(buf)) 
	{
		win_edit();
	}
}

int substring(const char *haystack, const char *needle, int ic)
{
	const char *h, *n;

	for ( ; ; haystack++ ) {
		for (h = haystack, n = needle; ; h++, n++) {
			if (*n == '\0')
				return 1;
			if (*h == '\0')
				return 0;
			if (ic ? tolower((unsigned char)*h) != tolower((unsigned char)*n)
			  : *h != *n)
				break;
		}
	}
	/* NOTREACHED */
	return 0;
}

static void
alloc_fail(size_t size)
{
	err_exit("Memory allocation failed, could not allocate %lu bytes",
	  (unsigned long)size);
	/* NOTREACHED */
}

/* malloc with error checking */
void *
emalloc(size_t size)
{
	void *mem;

	if (size > SSIZE_MAX)
		/*
		 * possible problems with signed/unsigned int !
		 *
		 * It is not normal to request such a huge memory
		 * block anyway (16-bit systems are not supported)
		 */
		alloc_fail(size);
	if ((mem = malloc(size)) == 0)
		alloc_fail(size);
	return mem;
}

/* realloc with error checking */
void * erealloc(void *ptr, size_t size)
{
	void *mem;

	/* not sure if really all realloc()s can handle this case */
	if (ptr == 0)
		return emalloc(size);

	if (size > SSIZE_MAX)
		/* see emalloc() above */
		alloc_fail(size);

	if ((mem = realloc(ptr,size)) == 0)
		alloc_fail(size);
	return mem;
}

/* strdup with error checking */
char * 
estrdup(const char *str)
{
	char *dup;

	if (str == 0)
		return 0;
	dup = emalloc(strlen(str) + 1);
	strcpy(dup,str);
	return dup;
}

/* set the directory name for pathname_join() */
void
pathname_set_directory(const char *dir)
{
	char *str;

	sf_dirlen = strlen(dir);
	us_resize(&sf_buff,sf_dirlen + 24);
	/* 24 extra bytes = slash + initial space for the filename */
	str = USTR(sf_buff);
	strcpy(str,dir);
	if (str[sf_dirlen - 1] != '/')
		str[sf_dirlen++] = '/';
		/* the string is not null terminated now, that's ok */
}

/*
 * join the filename 'file' with the directory set by
 * pathname_set_directory() above
 *
 * returned data is overwritten by subsequent calls
 */
char *
pathname_join(const char *file)
{
	us_resize(&sf_buff,sf_dirlen + strlen(file) + 1);
	strcpy(USTR(sf_buff) + sf_dirlen,file);
	return USTR(sf_buff);
}


/*
 * dequote backslash quoted text: 'src' -> 'dst'
 * 'dst' buffer must be large enough
 */
size_t
dequote_txt(const char *src, size_t len, char *dst)
{
	char ch;
	size_t i, j;
	FLAG quote;

	for (i = j = 0, quote = 0; i < len; i++) {
		ch = src[i];
		if (TCLR(quote) || !(quote = ch == '\\'))
			dst[j++] = ch;
	}
	dst[j] = '\0';

	return j;
}

/*
 * under certain condition can read() return fewer bytes than requested,
 * this wrapper function handles it
 *
 * remember: error check should be (read() == -1) and not (read() < 0)
 */
ssize_t
read_fd(int fd, char *buff, size_t bytes)
{
	size_t total;
	ssize_t rd;

	for (total = 0; bytes > 0; total += rd, bytes -= rd) {
		rd = read(fd,buff + total,bytes);
		if (rd == -1)	/* error */
			return -1;
		if (rd == 0)	/* EOF */
			break;
	}
	return total;
}

